/*-----------------------------------------------------------------
 * lfn.cpp : Longfilenames for DOS
 * compile in LARGE modell (FAR pointers)
 *
 * Copyright (C) 1999 Ortwin Glueck
 *
 * GNU General Public License applies
 *
 *
 * 2002 Modified by Herbert Reichardt
 * for  Borland C 3.1, Cache and CD files completely removed
 * designed for use with LBACK.CPP
 *
 * 2002 Modified by Sam Hawker <shawkie@geocities.com>
 * Reinstated Odi's makeAlias function
 * Added in Odi's makeDirectory function and some supporting functions
 * (for use with lunzip.cpp)
 *
 *-----------------------------------------------------------------
 */

//Define language
#define LANG_EN
//#define LANG_DE

//add Testcase TESTDEBUG for testing with a imagefile "test.ima"
//for functions ReadSectorP WriteSectorP
//lockDrive unlockDrive



// includes
#include <stdio.h>
#include <dos.h>
#include <conio.h>
#include <direct.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#ifdef TESTDEBUG

#include <io.h>
#include <fcntl.h>

#endif
// #include <vmemory.h>
#include "lfn.h"
#include "unicode.inc"
#include "wstring.cpp"

#ifdef LANG_DE
 #include "lfnmsgde.inc"
#endif
#ifdef LANG_EN
 #include "lfnmsgen.inc"
#endif

//Functions

//#pragma optimize("gel",off) //cannot optimize asm

void getDOSver(struct dosver *v)
{
	union REGS Regs;
	Regs.x.ax = 0x3000;          /* get version   */
	intdos(&Regs,&Regs);
	v->main=Regs.h.al;
	v->sub=Regs.h.ah;
	v->oem=Regs.h.bh;
}


int directAccessDenied(char drive) {
	//tell if drive is remote and if we can make direct i/o
	//0: allowed
	//1: notallowed
	int att,e;
	union REGS Regs;

	e=1;                         /* preset error */

	Regs.x.ax = 0x4409;          /* get version   */
	Regs.h.bl = drive+1;
	intdos(&Regs,&Regs);

	att = Regs.x.dx;
	if (!Regs.x.cflag)
		e = 0;

	if (e == 1) {
		return 1; //error
	}
	else {
		return (att & 0x8100); //bit 9 prohibits direct i/o, bit 15 is a subst
	}
}


#ifndef TESTDEBUG

int readSectorP(char drive,dword sector,word count, void *buffer) //extended version
//Read a physical sector (uncached)
{
	struct secblk x;
	int cy = 1;             /* preset error */
	union REGS Regs;
	struct SREGS SRegs;

	x.sector=sector;
	x.no_sectors=count;
	x.pBuffer=buffer;


	if (win98 == 1) // ---------- use fat32 compatible read
	{
		Regs.x.si = 0;              // read
		Regs.x.cx = 0xffff;
		Regs.x.bx = FP_OFF(&x);    /* DS:BX = ptr -> disk I/O Packet x  */
		SRegs.ds  = FP_SEG(&x);
		Regs.h.dl = drive+1;

		Regs.x.ax = 0x7305;          /* fat32 compatible read   */
		intdosx(&Regs,&Regs,&SRegs); /* call */

		cy = Regs.x.ax;
		if (!Regs.x.cflag)
			cy = 0;

		if (cy!=0) {
			if (cy == 1)
				win98 = 0; //function not supported, try normal read
			else {
				printf(msg[0],drive, sector, sector + count -1 );
				return 1;
			}
		}
		else return 0;
	}

	//use normal read, also if first fails. (NO ELSE)
	{
			asm push ds              /* save ds */
			asm mov  al,[drive]      /* 0=A:,1=B: usw.     */
			asm push ss              /* DS:BX = @ParmBlock x */
			asm pop  ds              /* (auf dem Stack)    */
			asm lea  bx,[x]          /* (Offset-Anteil)    */
			asm mov  cx,0xFFFF       /* Flag fuer ParmBlock */
			asm int  0x25            /* Aufruf             */
			asm pop  cx              /* STACK-KORREKTUR    */
			asm jc   error           /* jump if -> Fehler  */
			asm xor  al,al           /* sonst Status = 0   */
			asm xor  ah,ah           /* AH = undefiniert   */
error:
			asm pop  ds              /* restore ds         */
			asm mov  [cy],ax         /* Status setzen      */
	}

	if (cy != 0) {
		printf(msg[1],drive,sector,sector+count-1);
		return 1;
	}
	else return 0;
}



int writeSectorP(char drive,dword sector,word count, void *buffer,word typ) //extended version
//write a physical sector (uncached)
//type: 0x2000 =FAT, 0x4000 =DIR, 0x6000 =FILE
{
	struct secblk x;
	int cy = 1;                   /* preset error */
	union REGS Regs;
	struct SREGS SRegs;

	x.sector=sector;
	x.no_sectors=count;
	x.pBuffer=buffer;

	if (win98 == 1) // ---------- use fat32 compatible write
	{
		Regs.x.si = typ | 0x0001;  // type | write
		Regs.x.cx = 0xffff;
		Regs.x.bx = FP_OFF(&x);    /* DS:BX = ptr -> disk I/O Packet x  */
		SRegs.ds  = FP_SEG(&x);
		Regs.h.dl = drive+1;

		Regs.x.ax = 0x7305;          /* fat32 compatible r/w   */
		intdosx(&Regs,&Regs,&SRegs); /* call */

		cy = Regs.x.ax;
		if (!Regs.x.cflag)
			cy = 0;

		if (cy!=0) {
			if (cy==1)
				win98 = 0; //function not supported, try standard write
			else {
				printf(msg[2],drive,sector,sector+count-1);
				return 1;
			}
		}
		else return 0;
	}

	{                           //------------ try standard write ------
			asm push ds              /* save ds */
			asm mov  al,[drive]      /* 0=A:,1=B: usw.     */
			asm push ss              /* DS:BX = @ParmBlock x */
			asm pop  ds              /* (auf dem Stack)    */
			asm lea  bx,[x]          /* (Offset-Anteil)    */
			asm mov  cx,0xFFFF       /* Flag fuer ParmBlock */
			asm int  0x26            /* Aufruf             */
			asm pop  cx              /* STACK-KORREKTUR    */
			asm jc   werror           /* jump if -> Fehler  */
			asm xor  al,al           /* sonst Status = 0   */
			asm xor  ah,ah           /* AH = undefiniert   */
	werror:
			asm pop  ds              /* restore ds         */
			asm mov  [cy],ax         /* Status setzen      */
	}

	if (cy != 0) {
		printf(msg[3],drive,sector,sector+count-1);
		return 1;
	}

	return 0;
}

#endif


int readSector(char drive,dword sector,word count, void *buffer)
//read a cached sector
{
#ifndef TESTDEBUG
	return readSectorP(drive,sector,count,buffer);
#else
	lseek(debugpath,sector*512,0);
	read(debugpath,buffer, count*512);
	return 0;
#endif


}

int writeSector(char drive,dword sector,word count, void *buffer,word typ)
//write a cached sector
{
#ifndef TESTDEBUG
	return writeSectorP(drive,sector,count,buffer,typ);
#else
	lseek(debugpath,sector*512,0);
	write(debugpath,buffer, count*512);
	return 0;
#endif

}


// -------------- lock volume calls -------------------------------------

int LockVolumeCall(unsigned int uiLevel, byte cDrive)
{
	union REGS sInRegs;

	sInRegs.x.ax = 0x440d;			// Function-Number
	sInRegs.h.bh = (byte)uiLevel;	        // Lock-Level
	sInRegs.h.bl = cDrive;	        	// drive
	sInRegs.h.ch = 0x08;			// device category (must be 08h)
	sInRegs.h.cl = 0x4A;                    // Minor Code: "Lock Logical Volume"
	sInRegs.x.dx = 0x0001;			// Write permission for OTHER Processes (maybe better 0 ?)

	intdos (&sInRegs, &sInRegs);

	if (!sInRegs.x.cflag)
	   return 0;
	else
		return 1;

}


int UnlockVolumeCall(char cDrive)
{
	union REGS sInRegs;

	sInRegs.x.ax = 0x440d;		      // Function-Number
	sInRegs.h.bl = cDrive;		      // Laufwerk
	sInRegs.h.ch = 0x08;		      // device category (must be 08h)
	sInRegs.h.cl = 0x6A;                  // Minor Code: "Unlock Logical Volume"

	intdos (&sInRegs, &sInRegs);

	if (!sInRegs.x.cflag)
		return 0;                           // o.k.
	else
		return 1;
}



int lockDriveW(byte drv)
{
	int cy;
	if (noLocking==1)
		  return 0; //not supported


	cy = LockVolumeCall(1, drv+1);     // Locklevel 1 = Locked gerade andere Task?
	if (!cy)
		  cy = LockVolumeCall(2, drv+1); // Locklevel 2 = Schreiben jetzt einstellen
	else {
		  UnlockVolumeCall(drv+1);      // Zurueck von 1 in Ungelockten Zustand
		  goto LockErr;
	}
	cy = LockVolumeCall(3, drv+1);    // Locklevel 3 = jetzt schreiben wir
	if (cy) {
		UnlockVolumeCall(drv+1);       // Back von 2 auf Level 1
		UnlockVolumeCall(drv+1);       // Back von 1 in Ungelockten Zustand
	}

LockErr:
	if (cy == 1) {
		printf(msg[5],drv+'A');
		exit(1);
	}
	return 0;
}



int unlockDriveW(byte drv)
{

	int cy;

	if (noLocking==1)
		return 0;               //not supported

	UnlockVolumeCall(drv+1);	// Back from 3 to Level 2
	UnlockVolumeCall(drv+1);        // Back from 2 to Level 1
	cy = UnlockVolumeCall(drv+1);   // Back from 1 to unlocked

	if (cy==1) {
		printf(msg[6],drv);
		return 1;
	}
	return 0;
}



int lockDrive(byte drv)
{
	int cy;
	union REGS Regs;

#ifdef TESTDEBUG
	return 0;
#endif

	if (winactive == 1)
		return lockDriveW(drv);

	if (noLocking==1)
		return 0; //not supported

	Regs.x.ax = 0x440D;          /* lock/unlock drive   */
	Regs.h.bh = 0;
	Regs.h.bl = drv+1;
	Regs.x.cx = 0x084A;
	Regs.x.dx = 0x0000;

	intdos(&Regs,&Regs);
	if (!Regs.x.cflag)
		cy = 0;
	else
		cy = 1;

	if (cy == 1) {
		printf(msg[5],drv+'A');
		exit(1);
	}
	return 0;
}


int unlockDrive(byte drv)
{
	int cy;
	union REGS Regs;

#ifdef TESTDEBUG
	return 0;
#endif


	if (winactive == 1)
		return unlockDriveW(drv);

	if (noLocking==1)
		return 0;               //not supported

	Regs.x.ax = 0x440D;          /* lock/unlock drive   */
	Regs.h.bh = 0;
	Regs.h.bl = drv+1;
	Regs.x.cx = 0x086A;         //device category (must be 08h)
	Regs.x.dx = 0x0000;

	intdos(&Regs,&Regs);
	if (!Regs.x.cflag)
		cy = 0;
	else
		cy = 1;

	if (cy==1) {
		printf(msg[6],drv);
		return 1;
	}
	return 0;
}


int windows_active( void )
{
 union  REGS  regs;
// return bh major bl minor version number
 regs.x.ax = 0x160A;
 int86(0x2F, &regs, &regs);
 if ((regs.x.ax == 0) && (!regs.x.cflag))
	 return 1;
 else
	 return 0;
}


//---------------- first call this before anything else -------------
int makeDPB(char drive,struct DPB &d)
//Build Disk Parameter Block
{
	int plausible=1;
	char *sector=new char[MAX_SECTOR_SIZE];
	struct BPB *b=(struct BPB*)&sector[0x0B];
	struct BPB32 *b32=(struct BPB32*)&sector[0x0B];
	struct dosver ver;
	// struct iso_primary_descriptor    D;

	getDOSver(&ver);
	if (ver.oem!=0xff) {
		win98 = 0;  //PC-DOS does not support this!
		noLocking=1;
	}
	if ((100*ver.main+ver.sub)<710) win98 = 0; //is it >=Windows 98?
	if (ver.main<7) noLocking=1;  //old DOS does not support locking

	if (windows_active())
		winactive = 1;
	else
		winactive = 0;

	if (directAccessDenied(drive)!=0) {
		printf(msg[51]);
		exit(1);
	}

	// no cd support
	d.CD=0;
	d.CD_vol_start=0;
	d.cdfs=(CDFS_T)-1;

	if (readSector(drive,0,1,sector)) {
		printf(msg[16]);
		delete sector;
		return 1;
	}

	//Let the sector tell us what fs it is
	if (!strncmp("FAT12",b->fat_type,5)) d.fat_type=FAT12;
	else if (!strncmp("FAT16",b->fat_type,5)) d.fat_type=FAT16;
	else if (!strncmp("FAT32",b32->fat_type,5)) d.fat_type=FAT32;
	else d.fat_type=NO; //not stated in the sector, we have to decide later

	d.drive=drive;
	d.next_cluster=2;
	d.bytes_per_sector=b->bytes_per_sector;

	//bpsd[drive]=d.bytes_per_sector;

	d.sectors_per_cluster=b->sectors_per_cluster;
	d.reserved_sectors=b->reserved_sectors;
	d.no_fats=b->no_fats;
	d.no_root_entries=b->no_root_entries;
	d.media_descriptor=b->media_descriptor;
	if (b->sectors_per_fat==0)
		d.sectors_per_fat=b32->sectors_per_fat32;
	else
		d.sectors_per_fat=b->sectors_per_fat;
	if (b->no_sectors==0) d.no_sectors=b->no_large_sectors;
	else d.no_sectors=b->no_sectors;

	//plausibility check
	if ((d.bytes_per_sector<256) || (d.bytes_per_sector>=65535U) || (d.bytes_per_sector%256!=0) ||
		(d.sectors_per_cluster>=255) || (d.sectors_per_cluster==0) ||
		(d.no_fats>8)) {
		printf(msg[50]);
		plausible=0;
	}

	//FAT Type not yet clear
	if (d.fat_type==NO) {
		if (d.no_sectors/d.sectors_per_cluster>268435456L) {
			printf(msg[17]);
			delete sector;

			exit(1);
		}
		if (d.no_sectors/d.sectors_per_cluster<4087) d.fat_type=FAT12;
		else if (d.no_sectors/d.sectors_per_cluster>65525L) d.fat_type=FAT32;
		else d.fat_type=FAT16;
	}

	if ((d.fat_type==FAT12) || (d.fat_type==FAT16)) {
		d.no_root_sectors=(word)d.no_root_entries*32/d.bytes_per_sector;
		d.first_root_sector=(dword)d.reserved_sectors+(dword)d.no_fats*d.sectors_per_fat;
		d.first_data_sector=(dword)d.first_root_sector+(dword)d.no_root_sectors;
		d.first_root_cluster=0;
		strnset(d.label,0,sizeof(d.label));
		strncpy(d.label,b->volume_label,11);
	}
	else //FAT32
	{
		d.no_root_sectors=0;
		d.first_root_cluster=b32->cluster_of_root;
		d.first_data_sector=(dword)d.reserved_sectors+(dword)d.no_fats*d.sectors_per_fat;
		d.fs_info_sector=b32->fs_info_sector;
		d.first_root_sector=d.first_data_sector+(dword)d.sectors_per_cluster*(d.first_root_cluster-2);
		strncpy(d.label,b->volume_label,11);

		//read FSInfo
		struct BIGFATBOOTFSINFO *fsi;
		fsi=(struct BIGFATBOOTFSINFO*)malloc(b->bytes_per_sector);
		if (fsi==NULL) {
			printf(msg[27]);
			return 1;
		}
		readSector(d.drive,d.fs_info_sector,1,fsi);
		if ((fsi->start_signature==0x41615252L) && (fsi->signature==0x61417272L) && (fsi->end_signature==0xAA550000L)) {
			d.next_cluster=fsi->next_free_cluster;
		}
		free(fsi);
	}
	d.CD_iso_root=d.first_root_sector;
	d.no_clusters=(d.no_sectors-d.first_data_sector)/d.sectors_per_cluster+2; //including 2 dummy clusters at beginning of FAT

	switch (d.fat_type) {
	case FAT12: d.fat_entries_per_sector=d.bytes_per_sector/3*2; break;  //may vary +/- 1 !!! Do not use!
	case FAT16: d.fat_entries_per_sector=d.bytes_per_sector/2;   break;
	case FAT32: d.fat_entries_per_sector=d.bytes_per_sector/4;
		d.no_root_entries=0;
		break;
	}
	d.dir_entries_per_sector=d.bytes_per_sector/sizeof(struct direntry);
	delete sector;
	if (plausible==0) return 1;
	return 0;
}

byte makeChecksum(char *aliasname)
{
	byte sum,i;
	for (sum = i = 0; i < 11; i++)
		sum = (((sum&1)<<7)|((sum&0xfe)>>1)) + aliasname[i];
	sum &= 0xff;
	return sum;
}

void unincpy(word *dest,word *source,word count)
{
	word i;
	for (i=0;i<count;i++) dest[i]=source[i];
}

void lfnString(struct lfnentry &lfn,char *bs,word *ws) // 13 characters plus \0
{
	unincpy(&ws[0],&lfn.name1[0],5);
	unincpy(&ws[5],&lfn.name2[0],6);
	unincpy(&ws[11],&lfn.name3[0],2);
	ws[13]=0;

	uni2ascii((unsigned char*)ws,(unsigned char*)&bs[0],0);  // change here xxxx
}

unsigned int strpos(const char *string,int len,char c)
{
	// int i;

	for (int i=0;i<len;i++) {
		if (string[i]==c) break;
		if (string[i]==0) break;
	}
	return i;
}




////////////
//////////// Functions making the whole thing independent of FAT type
////////////

dword FATsector(const struct DPB *b,dword cluster)
//returns sector containing the FAT entry
{
	dword sector;

	switch (b->fat_type) {
	case FAT12:
		sector = cluster*3/2/b->bytes_per_sector+b->reserved_sectors; //returns sector containing the first byte
		break;
	case FAT16:
		sector = cluster/b->fat_entries_per_sector+b->reserved_sectors;
		break;
	case FAT32:
		sector = cluster/b->fat_entries_per_sector+b->reserved_sectors;
		break;
	default:
		sector = 0L;  //illegal
		break;

	}

	return sector;
}

#pragma warn -sig             // Warnung abschalten

dword getFATval(const struct DPB *b,void *FAT,dword cluster)
{
	dword pos;
	word val;
	dword posret;

	switch (b->fat_type) {
	case FAT12:
		pos=cluster/2*3; //abs position of triple byte
		if (cluster%2==0) {
			pos = pos % b->bytes_per_sector;
			val=((byte *)FAT)[pos];
			if (pos+1 < b->bytes_per_sector)
				val += ((((byte*)FAT)[pos+1] & 0x0f) << 8); //even Cluster:  bc 0a 00
			else
				fat12of=1; //entry is in both sectors
		}
		else {
			pos++;
			pos=pos % b->bytes_per_sector;
			val=((((byte*)FAT)[pos] & 0xf0) >> 4);
			if (pos+1<b->bytes_per_sector)
				val+=(((byte*)FAT)[pos+1] << 4); //odd Cluster: 00 c0 ab
			else
				fat12of=1; //entry is in both sectors
		}
		posret = (dword)val;
		break;
	case FAT16:
		pos=cluster%b->fat_entries_per_sector;
		posret = (dword)((word*)FAT)[pos];
		break;
	case FAT32:
		pos=cluster % b->fat_entries_per_sector;
		posret = (dword)((dword*)FAT)[pos];      //highes 4 bits are reserved and not part of the cluster number!
		break;

	default:
		posret = 0L; // illegal
		break;
	}
	return posret;
}

#pragma warn .sig             // Warnung wieder ein

#pragma warn -sig             // Warnung abschalten

void setFATval(const struct DPB *b,void *FAT,dword cluster,dword val)
{
	dword pos;
	switch (b->fat_type) {
	case FAT12:
		if ((word)val==0xffff) val=0xfff;
		pos=cluster/2*3;  //abs position of triple byte
		if (cluster%2==0) //even: bc 0a 00
		{
			pos=pos % b->bytes_per_sector;
			((byte*)FAT)[pos]=(byte)(val & 0xff);
			if (pos+1<b->bytes_per_sector) {
				((byte*)FAT)[pos+1]&=0xf0;
				((byte*)FAT)[pos+1]|=(byte)((val & 0x0f00) >> 8);
			}
			else
				fat12of=1; //entry is in both sectors
		}
		else               //odd: 00 c0 ab
		{
			pos++;
			pos=pos % b->bytes_per_sector;
			((byte*)FAT)[pos]&=0x0f;
			((byte*)FAT)[pos]|=(byte)((val & 0x00f) << 4);
			if (pos+1<b->bytes_per_sector)
				((byte*)FAT)[pos+1]=(byte)((val & 0xff0) >> 4);
			else
				fat12of=1; //entry is in both sectors
		}
		break;
	case FAT16:
		pos=cluster % b->fat_entries_per_sector;
		((word*)FAT)[pos]=(word)val;
		break;
	case FAT32:
		pos=cluster % b->fat_entries_per_sector;
		((dword*)FAT)[pos]=val;
		break;
	}
}

#pragma warn .sig             // Warnung wieder ein

dword getFAT12valX(const struct DPB *b,dword cluster,dword val)
//get the rest of the FAT entry when fat12of occured
{
	void *FAT=malloc(b->bytes_per_sector);

	if (FAT==NULL) {
		printf(msg[18]);

		exit(1);
	}
	readSector(b->drive,FATsector(b,cluster)+1,1,FAT);

	if (cluster % 2==0)
		val+=((((byte*)FAT)[0] & 0x0f) << 8); //even Cluster:  bc 0a 00
	else
		val+=(((byte*)FAT)[0] << 4); //odd Cluster: 00 c0 ab

	free(FAT);
	fat12of=0;
	return val;
}

void setFAT12valX(const struct DPB *b,dword cluster,dword val)
{
	void *FAT=malloc(b->bytes_per_sector);
	dword sector;

	if (FAT==NULL) {
		printf(msg[19],'A'+b->drive,cluster);

		exit(1);
	}
	sector=FATsector(b,cluster)+1;
	readSector(b->drive,sector,1,FAT);

	if (cluster % 2==0) {
		((byte*)FAT)[0]&=0xf0;
		((byte*)FAT)[0]|=(byte)((val & 0x0f00) >> 8);
	}
	else {
		((byte*)FAT)[0]=(byte)((val & 0xff0) >> 4);
	}
	for (int fats=0;fats<b->no_fats;fats++)
		writeSector(b->drive,sector+fats*b->sectors_per_fat,1,FAT,W_FAT);     //update all fats

	free(FAT);
	fat12of=0;
}

dword getNextAvailClusterNo(struct DPB *b) //use 1st fat
{
	void* FAT=NULL;
	dword sector,current_sector;
	dword total,val;

	FAT=(word*)malloc(b->bytes_per_sector);
	if (FAT==NULL) {
		printf(msg[20]);

		exit(1);
	}

	if ((b->fat_type==FAT16) || (b->fat_type==FAT12) || (b->fat_type==FAT32)) {
		do {
			current_sector=0; //invalid value so 1st time sector is read always
			for (total=b->next_cluster;total<b->no_clusters;total++) {
				sector=FATsector(b,total);
				if (sector!=current_sector) {
					readSector(b->drive,sector,1,FAT);
					current_sector=sector;
				}
				val=getFATval(b,FAT,total);
				if (fat12of==1) val=getFAT12valX(b,total,val);
				if ((val & FAT_MASK)== 0L) {
					free(FAT);
					b->next_cluster=total;
					if (total==b->no_clusters-1) b->next_cluster=2;
					return total;
				}
			}
			if (b->next_cluster==2) break;
			else b->next_cluster=2;
		} while (1==1); //if not found the 1st time, try once from beginning of disk.
		free(FAT);
	}
	else {
		free(FAT);
		printf(msg[21]);

		exit(1);
	}
	return 0;
}

dword firstSectorOfCluster(const struct DPB *b,dword cluster)
{
	if (cluster==0) return b->first_root_sector; //root
	if (b->CD==1) return cluster;
	return b->first_data_sector+(dword)b->sectors_per_cluster*(cluster-2);
}

int getNextCluster(const struct DPB *b,struct FATfile *f)
//search for next FAT entry starting with f
//if f is the last cluster in a chain or free, f is unchanged and return is 1
{
	word *FAT;
	dword FATval;
	dword sector;

	if ((b->fat_type==FAT16) || (b->fat_type==FAT12) || (b->fat_type==FAT32)) {
		FAT=(word*)malloc(b->bytes_per_sector);
		if (FAT==NULL) {
			printf(msg[22]);

			exit(1);
		}
		sector=FATsector(b,f->cluster);
		readSector(b->drive,sector,1,FAT);
		FATval=getFATval(b,FAT,f->cluster);
		if (fat12of==1) FATval=getFAT12valX(b,f->cluster,FATval);
		FATval &= FAT_MASK;
		free(FAT);
		if (((b->fat_type==FAT16) && (FATval>=EOF16)) ||  //EOF?
			((b->fat_type==FAT12) && (FATval>=EOF12 )) ||
			((b->fat_type==FAT32) && (FATval>=EOF32))) {
			return 1;
		}
		else {
			if (FATval==0) {
				return 1; //a hole in a chain or no chain
			}
			f->cluster=FATval; //continue in chain
			f->sector=firstSectorOfCluster(b,FATval);
			return 0;
		}
	} // if
	else {
		printf(msg[23]);

		exit(1);
	}
	return 0;
}

char* shortname(const struct longdirentry* d) //returned string is static
{
	static char s[13];     // wichtig static !!!
	int i,j;
	for (i=7; i > 0 && (d->dosname[i]== ' ');) i--;  // search end of padding from right side
	strncpy(s,d->dosname,i+1);
	j=i+1;
	s[j] = 0;
	for (i=2; i >= 0 && (d->dosext[i]== ' ');) i--;  // search end of padding from right side
	if (i < 0)
		return s;
	s[j++] = '.';
	strncpy(&s[j],d->dosext,i+1);
	s[j+i+1] = 0;
	return s;

#if 0
	for (i=0;(d->dosname[i]!=' ') && (i<8);i++) s[i]=d->dosname[i];
	if (d->dosext[0]!=' ') {
		s[i]='.';
		i++;
		for (j=0;(d->dosext[j]!=' ') && (j<3);j++) { s[i]=d->dosext[j];i++;
		}
	}
	s[i]=0;
	return s;
#endif

}

int match(const char *name,const char *mask) // match with upper lower case ignore
{
	if (!strcmp("*",mask)) return 1;
	return !strcmpi(name,mask);
}

int matchexact(const char *name,const char *mask) // match exact
{
	if (!strcmp("*",mask)) return 1;          // * is always match
	return !strcmp(name,mask);                // avoid any codepage dependency
}

int WildComp(const char *I,const char *W) //match with wildcard search
{
	char subs[255],WildS[255],IstS[255];
	byte i, j, l, p;
	char *ps;
	strcpy(WildS,W);
	strcpy(IstS,I);
	strupr(IstS);
	strupr(WildS);
	i = 0;
	j = 0;
	while (i<=strlen(WildS)) {
		if (WildS[i]=='*') {
			if (i == strlen(WildS)) return 1; //found
			else // we need to synchronize
			{
				l = i+1;
				while ((l < strlen(WildS)) && (WildS[l+1] != '*')) l++;
				strncpy(&subs[0],&WildS[i+1], l-i);
				subs[l-i]=0;
				if (subs[0]==0) return 1;
				ps= strstr(IstS,subs);
				if (ps!=NULL) {
					p = (byte)(ps-IstS);
					j = p-1;
				}
				else return 0;
			}
		}
		else
			if ((WildS[i]!='?') && ((strlen(IstS) < i) || (asciiUpcase[(unsigned char)WildS[i]]!=asciiUpcase[(unsigned char)IstS[j]]))) return 0;
		i++;
		j++;
	}
	return (j > strlen(IstS));
}

void catUniStrg(word *dest,const word *source)
{
	int i,j;
	for (i=0;dest[i]!=0;i++); //seek to end of source string
	for (j=0;source[j]!=0;j++) dest[i+j]=source[j]; //copy
	dest[i+j]=0; //end of string mark
}

void catUniStrgn(word *dest, const word *source, const int count)
{
	int i,j,n;
	for (i=0;dest[i]!=0;i++); //seek to end of source string
	for (j=0, n=0;(n<count) && (source[j]!=0);j++, n++) dest[i+j]=source[j]; //copy
	dest[i+j]=0; //end of string mark
}


struct longdirentry FindMask(const struct DPB *b,char far *mask,struct dirpointer* start,int exact, struct direntry *pd)
//for fat12/fat16-rootdir: set start->cluster=0
//for fat32-rootdir: treat like a subdir
//wildcards allowed, LFN ONLY,
//mask: null or "" mask is *
//exact:bit 0 set: no wildcards allowed
//      bit 1 set: use short names
//      bit 2 set: use exact short name, no stricmp
{
	struct longdirentry pDir;
	struct direntry *dir;
	struct lfnentry *lfn;
	struct FATfile f;
	void *data;
	word entries_per_sector,i;
	char longfilename[MAX_PATH_SIZE],name[MAX_PATH_SIZE],cname[MAX_PATH_SIZE];
	word *longuniname,*uniname;
	byte checksum;
	int chain_end;

	pDir.name[0]=0;
	pDir.dosname[0]=0;
	pDir.dosext[0]=0;
	if (mask==NULL) mask= (char far *)&"*";
	if (mask[0]==0) mask="*";
	if (start==NULL) return pDir; //ungltiger Pointer

	data=malloc(b->bytes_per_sector);
	if (data==NULL) {
		printf(msg[24]);

		exit(1);
	}
	entries_per_sector=b->bytes_per_sector/32;
	if ((b->fat_type==FAT32) && (start->cluster==0)) start->cluster=b->first_root_cluster;
	f.cluster=start->cluster;

	if (f.cluster==0) //root?
	{
		f.sector=b->first_root_sector;
	}
	else f.sector=firstSectorOfCluster(b,start->cluster);

	longuniname=(word*)malloc(MAX_LFN_SIZE*sizeof(word));
	uniname=(word*)malloc(MAX_LFN_SIZE*sizeof(word));
	if ((longuniname==NULL) || (uniname==NULL)) {
		printf(msg[25]);
		free(data);

		exit(1);
	}
	longfilename[0]=0;
	longuniname[0]=0;
	checksum=0;
	do {
		do {
			if (readSector(b->drive,f.sector+start->relsector,1,data)) {
				printf(msg[26]);
				free(longuniname);
				free(uniname);
				free(data);

				exit(1);
			};
			dir=&((struct direntry*)data)[start->entry_no];
			lfn=&((struct lfnentry*)data)[start->entry_no];
			for (i=start->entry_no;i<entries_per_sector;i++) {
				start->entry_no++;
				if (dir->name[0]==0) //subdir end
				{
					pDir.name[0]=0;
					pDir.dosname[0]=0;
					pDir.dosext[0]=0;
					free(longuniname);
					free(uniname);
					free(data);
					return pDir;
				}
				if (dir->atrib!=F_LFN)            //short file name found
				{
					if (longfilename[0]!=0)
					   strcpy(pDir.name,longfilename); //name can be mangled due to Code Page conversion
					else
					   pDir.name[0]=0;
					pDir.uniname[0]=0;                // preset no unicode name
					if (longuniname[0]!=0) catUniStrg(pDir.uniname,longuniname); // copy exact unicode name
					if ((byte)dir->name[0]!=0xe5) //deleted?
					{
						strncpy(pDir.dosname,dir->name,8);
						strncpy(pDir.dosext,dir->ext,3);
						pDir.checksum=makeChecksum(pDir.dosname);
						if ((checksum!=0) && (checksum!=pDir.checksum)) {
							//checksums do not match, throw long name away.
							pDir.name[0]=0;
							pDir.uniname[0]=0;
						}                      // return short name
						if (pDir.name[0]==0) strcpy(pDir.name,shortname(&pDir));

								      // short or long name compare ?
						if ((exact & 2)==0)  strcpy(cname,pDir.name); //long
						else strcpy(cname,shortname(&pDir));          //short

						if ((((exact & 1) == 0) && WildComp(cname,mask))       // XX0: wildcard allowed
							 || (((exact & 5) == 1) && match(cname,mask))       // 0X1: no wildcard except * case insensitive
						    || (((exact & 5) == 5) && matchexact(cname,mask))) // 1X1: no wildcard except * case sensitive
						{
							pDir.ftype=0;
							pDir.time=dir->time;
							pDir.date=dir->date;
							pDir.start_cluster=dir->start_cluster+((dword)dir->high_cluster << 16);
							pDir.length=dir->length;
							pDir.atrib=dir->atrib;
							strncpy(pDir.reserved,dir->reserved,8);

							if (pd != NULL)                    // make exact copy of short dir entry
							   memcpy(pd, dir, sizeof(struct direntry));

							free(longuniname);
							free(uniname);
							free(data);
							return pDir;
						}                        // correct entry found >>>>>
					}
					longfilename[0]=0;
					longuniname[0]=0;
				}
				else  //long file name
				{
					if (lfn->count!=0xe5)   // deleted?
					{
						if ((checksum!=0) && (lfn->checksum!=checksum)) {
							//wrong checksum, start over
							longfilename[0]=0;
							name[0]=0;
							uniname[0]=0;
							longuniname[0]=0;
						}
						lfnString(*lfn,&name[0],uniname);
						strcat(name,longfilename);
						strcpy(longfilename,name);
						catUniStrg(uniname,longuniname);
						longuniname[0]=0;
						catUniStrg(longuniname,uniname);
						checksum=lfn->checksum;
					}
				}
				lfn++;
				dir++;
			} //for each entry
			start->entry_no=0;
			start->relsector++;
		} while (((f.cluster!=0) && (start->relsector<b->sectors_per_cluster)) ||
			((f.cluster==0) && (start->relsector<b->no_root_sectors)));
		if (f.cluster==0) {
			pDir.name[0]=0;
			pDir.uniname[0]=0;
			pDir.dosname[0]=0;
			pDir.dosext[0]=0;
			free(data);
			free(longuniname);
			free(uniname);
			return pDir; //root:not found
		}
		start->relsector=0;
		chain_end=getNextCluster(b,&f);
		start->cluster=f.cluster;
	} while (chain_end!=1);
	pDir.name[0]=0;
	pDir.uniname[0]=0;
	pDir.dosname[0]=0;
	pDir.dosext[0]=0;
	free(data);
	free(longuniname);
	free(uniname);
	return pDir; //not found
}

int updateFSInfo(const struct DPB *b,signed long allocated_clusters,dword last_cluster)
//FAT32 only
//allocated_clusters: relative number of newly allocated clusters, negative values are freed clusters
{
	struct BIGFATBOOTFSINFO *fsi;
	if (b->fat_type!=FAT32) return 1;
	if (allocated_clusters==0) return 0;
	fsi=(struct BIGFATBOOTFSINFO*)malloc(b->bytes_per_sector);
	if (fsi==NULL) {
		printf(msg[27]);

		exit(1);
	}
	readSector(b->drive,b->fs_info_sector,1,fsi);
	if ((fsi->start_signature==0x41615252L) && (fsi->signature==0x61417272L) && (fsi->end_signature==0xAA550000L)) {
		if (fsi->no_free_clusters != 0xFFFFFFFFL) {
			fsi->no_free_clusters-=allocated_clusters;
		}
		if (last_cluster!=0) fsi->next_free_cluster=last_cluster;
		writeSector(b->drive,b->fs_info_sector,1,fsi,W_OTHER);
	}
	free(fsi);
	return 0;
}

dword extendFile(struct DPB *b,struct FATfile* chain,dword count,int type)
//We implement a stupid disk cache to minimize (slow) direct disk access
//allocate additional count clusters in all fat copies for this file
//type=W_DIR: erase new clusters with 00
//return number of finally allocated clusters
{
	struct FATfile *last;
	dword clusterNo,recent;
	dword sector,fat_in_store,val;
	word *FAT,fats;
	word i,j;
	void *EMPTY=NULL;
	int dirty;

	if ((b->fat_type!=FAT16) && (b->fat_type!=FAT12) && (b->fat_type!=FAT32)) {
		printf(msg[28]);

		exit(1);
	}

	while (!getNextCluster(b,chain)) ; //seek end of chain
	last=chain;
	FAT=(word*)malloc(b->bytes_per_sector);
	EMPTY=malloc(b->bytes_per_sector);
	if ((FAT==NULL) || (EMPTY==NULL)) {
		printf(msg[29]);

		exit(1);
	}
	memset(EMPTY,0,b->bytes_per_sector);
	recent=0;
	fat_in_store=0;
	clusterNo=b->next_cluster-1;
	dirty=0;
	for (i=0;i<count;i++) {
		//clusterNo=next free cluster on drive
		do {
			do {
				clusterNo++;
				sector=FATsector(b,clusterNo);
				if (sector!=fat_in_store) {
					if (dirty==1) //WB dirty data
					{
						for (fats=0;fats<b->no_fats;fats++)
							writeSector(b->drive,fat_in_store+fats*b->sectors_per_fat,1,FAT,W_FAT); //update all fats
						dirty=0;
					}
					readSector(b->drive,sector,1,FAT); //get next fat sector
					fat_in_store=sector;
				}
				val=getFATval(b,FAT,clusterNo);
				if (fat12of==1) val=getFAT12valX(b,clusterNo,val);
			} while ((clusterNo<b->no_clusters) && ((val & FAT_MASK)!=0));
			if (b->next_cluster==2) break;
			else b->next_cluster=2;
		} while (clusterNo==b->no_clusters); //try again from beginning if optimized search did not succeed

		if (clusterNo>=b->no_clusters) //not found
		{
			if (dirty==1) //WB dirty data
			{
				for (fats=0;fats<b->no_fats;fats++)
					writeSector(b->drive,fat_in_store+fats*b->sectors_per_fat,1,FAT,W_FAT); //update all fats
				dirty=0;
			}
			free(EMPTY);
			free(FAT);
			updateFSInfo(b,i,recent);
			return i;  //disk full
		}
		//allocate it
		b->next_cluster=clusterNo;
		val=getFATval(b,FAT,clusterNo);
		if (fat12of==1) val=getFAT12valX(b,clusterNo,val);
		setFATval(b,FAT,clusterNo,(val & INV_FAT_MASK) | EOF32); //reserved bits are preserved
		if (fat12of==1) setFAT12valX(b,clusterNo,(val & INV_FAT_MASK)|EOF32);
		recent=clusterNo;
		dirty=1;

		//link to previous
		if (last->cluster!=clusterNo)  //only if there IS one to link
		{
			sector=FATsector(b,last->cluster);
			//cache
			if (sector!=fat_in_store) {
				if (dirty==1) //WB dirty data
				{
					for (fats=0;fats<b->no_fats;fats++)
						writeSector(b->drive,fat_in_store+fats*b->sectors_per_fat,1,FAT,W_FAT); //update all fats
					dirty=0;
				}
				readSector(b->drive,sector,1,FAT); //fetch new data
				fat_in_store=sector;
			}
			val=getFATval(b,FAT,last->cluster);
			if (fat12of==1) val=getFAT12valX(b,last->cluster,val);
			setFATval(b,FAT,last->cluster,(val & INV_FAT_MASK) | clusterNo); // link
			if (fat12of==1) setFAT12valX(b,last->cluster,(val & INV_FAT_MASK) | clusterNo);
			dirty=1;
		}
		last->cluster=clusterNo;
		last->sector=firstSectorOfCluster(b,clusterNo);
		//when appending a dir -> fill with zeros
		if (type==W_DIR) {
			for (j=0;j<b->sectors_per_cluster;j++)
				writeSector(b->drive,last->sector+j,1,EMPTY,type);
		}
	}
	if (dirty==1) //WB dirty data
	{
		for (fats=0;fats<b->no_fats;fats++)
			writeSector(b->drive,fat_in_store+fats*b->sectors_per_fat,1,FAT,W_FAT); //update all fats
		dirty=0;
	}
	free(FAT);
	free(EMPTY);
	recent = getNextAvailClusterNo(b);  //update to current
	updateFSInfo(b,i,recent);
	return i;
}


char far * seekPath(struct DPB *b,char *path,char * longpath,longdirentry &pDir,int exact)
//return the short path and direntry of object given by path
//user must free the space occupied by the returned string
//path: NULL: returns current directory; can handle drive letters
//exact bit 1==0: you can use wildcards, first occurence is passed back
//      bit 2==1: assume short names in path
{
	char far *spath;
	char far *shortret;

	char part[MAX_PATH_SIZE];
	char current [MAX_PATH_SIZE];
	// char s[MAX_PATH_SIZE];
	word z,k,z1;
	struct longdirentry f;
	struct dirpointer dp;

	if (path==NULL) {
		// size of current increased from MAX_LFN_SIZE to MAX_PATH_SIZE else crash !

		_getdcwd(b->drive+1,current,MAX_PATH_SIZE);
		if (current[strlen(current)-1]!='\\') strcat(current,"\\");

		shortret=seekPath(b,current,longpath,pDir,3);   // recursive
		return shortret;                                // note: free is done by caller
	}

	spath=(char far *)malloc(MAX_PATH_SIZE);
	if (spath==NULL) {

		printf(msg[30]);
		exit(1);

	}
	spath[0]='\0';
	longpath[0]='\0';
	z=0;
	if (path[1]==':') {
		strncat(spath,path,2);
		strncat(longpath,path,2);
		z=2;
	}
	strcat(spath,"\\");
	strcat(longpath,"\\");
	if (path[z]=='\\') // from root
	{
		z++;
		current[0]=0;
	}
	else //from current directory
	{
		_getdcwd(b->drive+1,current,MAX_PATH_SIZE);
		if (current[strlen(current)-1]!='\\') strcat(current,"\\");

		shortret = seekPath(b,current,longpath,pDir,3);
		if (shortret != NULL) free(shortret);

		dp.cluster=pDir.start_cluster;
		dp.relsector=0;
		dp.entry_no=0;
		dp.countfile=COUNTFILE_START;
		dp.countdir=COUNTDIR_START;
		dp.length=dp.length;
		z1=strlen(current);
		strcpy(spath,current);
		//  if ((b->CD==1) && (b->cdfs==CD_JOLIET)) strcpy(longpath,s);
		strcat(current,&path[z]);
		path=current;
		z=z1;
	}
	part[0]=0;
	k=strpos(&path[z],MAX_PATH_SIZE,'\\');
	strncat(part,&path[z],k); //without backslash
	z=z+k;
	f.name[0]=0;
	//FAT drive
	if (current[0]==0) {
		dp.cluster=b->first_root_cluster;
		dp.relsector=0;
		dp.entry_no=0;
		//  if (b->CD==1)
		//  {
		//	dp.countfile=COUNTFILE_START;
		//	dp.countdir=COUNTDIR_START;
		//	dp.length=b->no_root_entries;
		//  }
	}
	while (part[0]!=0) {
		f=FindMask(b,part,&dp,exact);
		if (f.name[0]==0) //found?
		{
			free(spath);
			return NULL; //not found
		}
		pDir=f;
		strcat(spath,shortname(&f));
		strcat(longpath,f.name);
		if (path[z]=='\\') {
			strcat(spath,"\\");
			strcat(longpath,"\\");
			z++;
		}
		if ((path[z]!=0) && ((f.atrib & F_DIR) ==F_DIR)) //subdir?
		{
			dp.cluster=f.start_cluster;
			dp.relsector=0;
			dp.entry_no=0;
			dp.countfile=COUNTFILE_START;
			dp.countdir=COUNTDIR_START;
			dp.length=f.length;
		}
		else {
			return _fstrupr(spath); //end
		}
		part[0]=0;
		k=strpos(&path[z],MAX_PATH_SIZE,'\\');
		strncat(part,&path[z],k); //without backslash
		z=z+k;
	} //while
	if ((f.name[0]==0) && (current[0]==0)) //it is the root dir
	{
		pDir.name[0]=0;
		pDir.uniname[0]=0;
		pDir.dosname[0]=0;
		pDir.dosext[0]=0;
		pDir.atrib=F_DIR;
		pDir.start_cluster=dp.cluster;
		if (b->CD==1)
			pDir.length=dp.length;
		else
			pDir.length=0;
	}
	return _fstrupr(spath);
}



void ChangeDir(struct DPB *b, char *path)
{
	char *shortpath=NULL;
	char longpath[MAX_PATH_SIZE];
	struct longdirentry dir;
	int l;

	shortpath=seekPath(b,path,longpath,dir,0);
	if (shortpath == NULL) {
		printf("Chgdir Error\n");
		return;
	}
	l=strlen(shortpath)-1;
	if ((l>2) && (shortpath[l]=='\\')) shortpath[l]='\0';
	_chdrive(b->drive+1);
	chdir(shortpath);
	if (shortpath != NULL) free(shortpath);
}

int drive_of_path(char *path)    //0=A, 1=B, ...
{
	unsigned drive;
	_dos_getdrive(&drive);
	if (path==NULL) return drive-1;
	if (path[1]==':') {

		return (toupper(path[0])-'A');
	}
	else return drive-1;
}


void deleteDirentry(struct DPB *b,struct longdirentry *dir,struct dirpointer *dp,struct longdirentry *f)
//dir: the directory to delete the entry from
//dp: position inside directory
//f: the direntry (name) to delete
{
	// struct direntry *d;
	struct lfnentry *l;
	struct FATfile fatf;
	dword  prev;
	void *data;
	word i;
	byte count,no_lfn;

	data=malloc(b->bytes_per_sector);
	if (data==NULL) {
		printf(msg[30]);

		exit(1);
	}
	// d=(struct direntry*)data;
	l=(struct lfnentry*)data;
	if (f->uniname[0]==0) //no lfn
		no_lfn=0;
	else //has lfn
	{
		no_lfn = Wstrlen(f->uniname)/CHR_PER_LFN; //how many lfns are used?
		if (Wstrlen(f->uniname)%CHR_PER_LFN!=0) no_lfn++;
	}
	count=no_lfn;
	if (no_lfn>dp->entry_no) //some are in a different sector
	{
		if (dp->relsector==0)  //lfn in previous cluster
		{
			if (dir->start_cluster==0) {
				printf(msg[31]);
				free(data);

				exit(1);
			}
			fatf.cluster=dir->start_cluster;
			fatf.sector=firstSectorOfCluster(b,fatf.cluster);
			do //seek the previous cluster of the dir
			{
				prev=fatf.cluster; //this is the previous cluster
				if (getNextCluster(b,&fatf)==1) {
					printf(msg[32]);
					free(data);

					exit(1);
				}
			} while (fatf.cluster!=dp->cluster);

			readSector(b->drive,firstSectorOfCluster(b,prev+1)-1,1,data); //read the last sector of the previous cluster
			for (i=b->dir_entries_per_sector-(no_lfn- dp->entry_no);i<b->dir_entries_per_sector;i++) {
				if ((i!=dp->entry_no) && ((l[i].count&0x3f)!=count)) //security check
				{
					printf(msg[33]);
					free(data);

					exit(1);
				}
				l[i].count=0xe5; //mark as deleted
				count--;
			}
			writeSector(b->drive,firstSectorOfCluster(b,prev+1)-1,1,data,W_DIR);
			no_lfn=(byte)dp->entry_no;
		}
		else //in the same cluster
		{
			readSector(b->drive,dp->relsector-1+firstSectorOfCluster(b,dp->cluster),1,data);
			for (i=b->dir_entries_per_sector-(no_lfn- dp->entry_no);i<b->dir_entries_per_sector;i++) {
				if ((i!=dp->entry_no) && ((l[i].count&0x3f)!=count)) //security check
				{
					printf(msg[34]);
					free(data);

					exit(1);
				}
				l[i].count=0xe5; //mark as deleted
				count--;
			}
			writeSector(b->drive,dp->relsector-1+firstSectorOfCluster(b,dp->cluster),1,data,W_DIR);
			no_lfn=(byte)dp->entry_no;
		}
	}
	//rest is in the same sector

	readSector(b->drive,dp->relsector+firstSectorOfCluster(b,dp->cluster),1,data);
	for (i=dp->entry_no-no_lfn;i<=dp->entry_no;i++) {
		if ((i!=dp->entry_no) && ((l[i].count&0x3f)!=count)) //security check
		{
			printf(msg[35]);
			free(data);

			exit(1);
		}
		l[i].count=0xe5; //mark as deleted
		count--;
	}
	writeSector(b->drive,dp->relsector+firstSectorOfCluster(b,dp->cluster),1,data,W_DIR);
	free(data);
}


// -------------------------------------------------

void RenameShortDirentryInPlace(const struct DPB *b,const struct longdirentry *dir,const struct dirpointer *dp,const struct longdirentry *f)

//dir: the directory to rename the entry from
//dp: position inside directory

//f: the direntry (name) to rename

//NOTE: only useful for renaming of short directory entries in place
//only checksum update for long directory name if any

{
	struct lfnentry *l;
	struct FATfile fatf;
	dword  prev;
	word   i;
	word   count,no_lfn;
	void *data;

	data=malloc(b->bytes_per_sector);
	if (data==NULL) {
		printf(msg[30]);

		exit(1);
	}

	l=(struct lfnentry*)data;
	if (f->uniname[0]==0) //no lfn
		no_lfn=0;
	else //has lfn
	{
		no_lfn = Wstrlen(f->uniname)/CHR_PER_LFN; //how many lfns are used?
		if (Wstrlen(f->uniname)%CHR_PER_LFN!=0) no_lfn++;
	}
	count=no_lfn;

	// ------- longfile checksum update if any lfns -----

	if (no_lfn>dp->entry_no) //some are in a different sector
	{
		if (dp->relsector==0)  //lfn in previous cluster
		{
			if (dir->start_cluster==0) {
				printf(msg[31]);
				free(data);

				exit(1);
			}
			fatf.cluster=dir->start_cluster;
			fatf.sector=firstSectorOfCluster(b,fatf.cluster);
			do //seek the previous cluster of the dir
			{
				prev=fatf.cluster; //this is the previous cluster
				if (getNextCluster(b,&fatf)==1) {
					printf(msg[32]);
					free(data);

					exit(1);
				}
			} while (fatf.cluster!=dp->cluster);

			readSector(b->drive,firstSectorOfCluster(b,prev+1)-1,1,data); //read the last sector of the previous cluster
			for (i=b->dir_entries_per_sector-(no_lfn- dp->entry_no);i<b->dir_entries_per_sector;i++) {
				if ((i!=dp->entry_no) && ((l[i].count&0x3f)!=count)) //security check
				{
					printf(msg[33]);
					free(data);

					exit(1);
				}
				if(count == 0)
				   memcpy((char *)&l[i].count,f->dosname, 11); // write new short name
				else
				   l[i].checksum=f->checksum; //update longfile checksum

				count--;
			}
			writeSector(b->drive,firstSectorOfCluster(b,prev+1)-1,1,data,W_DIR);
			no_lfn=(byte)dp->entry_no;
		}
		else //in the same cluster
		{
			readSector(b->drive,dp->relsector-1+firstSectorOfCluster(b,dp->cluster),1,data);
			for (i=b->dir_entries_per_sector-(no_lfn- dp->entry_no);i<b->dir_entries_per_sector;i++) {
				if ((i!=dp->entry_no) && ((l[i].count&0x3f)!=count)) //security check
				{
					printf(msg[34]);
					free(data);

					exit(1);
				}
				if(count == 0)
					memcpy((char *)&l[i].count,f->dosname, 11); // write new short name
				else
				   l[i].checksum=f->checksum; //update longfile checksum

				count--;
			}
			writeSector(b->drive,dp->relsector-1+firstSectorOfCluster(b,dp->cluster),1,data,W_DIR);
			no_lfn=(byte)dp->entry_no;
		}
	}
	//----- rest is in the same sector

	// --------- short name rename or lfn crc updates -----------------

	readSector(b->drive,dp->relsector+firstSectorOfCluster(b,dp->cluster),1,data);
	for (i=dp->entry_no-no_lfn;i<=dp->entry_no;i++) {
		if ((i!=dp->entry_no) && ((l[i].count&0x3f)!=count)) //security check
		{
			printf(msg[35]);
			free(data);

			exit(1);
		}
		if(count == 0)
			memcpy((char *)&l[i].count,f->dosname, 11); // write new short name
		else
		   l[i].checksum=f->checksum; //update longfile checksum

		count--;
	}
	writeSector(b->drive,dp->relsector+firstSectorOfCluster(b,dp->cluster),1,data,W_DIR);
	free(data);
}




int freeClusterChain(struct DPB *b,struct longdirentry *f)
{
	void* FAT;
	dword sector,cluster,temp,count;
	word fats;

	FAT=malloc(b->bytes_per_sector);
	if (FAT==NULL) {
		printf(msg[36]);

		exit(1);
	}
	count=0;
	cluster=f->start_cluster;
	if (cluster==0) {
		free(FAT);
		return 1;
	}
	if ((b->fat_type==FAT16) || (b->fat_type==FAT12) || (b->fat_type==FAT32)) {
		do {
			sector=FATsector(b,cluster);
			readSector(b->drive,sector,1,FAT);
			temp=getFATval(b,FAT,cluster); //save the entry value
			if (fat12of==1) temp=getFAT12valX(b,cluster,temp);
			if ((temp & FAT_MASK)==0) {
				printf(msg[37]);
				free(FAT);

				exit(1);
			}
			setFATval(b,FAT,cluster,0 | (temp & INV_FAT_MASK)); // free the cluster preserving reserved bits
			if (fat12of==1) setFAT12valX(b,cluster,0|(temp & INV_FAT_MASK));
			count++;
			for (fats=0;fats<b->no_fats;fats++)
				writeSector(b->drive,sector+fats*b->sectors_per_fat,1,FAT,W_FAT);     //update all fats
			cluster=(temp & FAT_MASK); //next one
		} while (((b->fat_type==FAT12) && (cluster<EOF12)) ||
			((b->fat_type==FAT16) && (cluster<EOF16)) ||
			((b->fat_type==FAT32) && (cluster<EOF32))); //until EOF
		if (b->fat_type==FAT32) updateFSInfo(b,-(signed long)count,0);
	}
	else {
		printf(msg[38]);
		free(FAT);

		exit(1);
	}
	free(FAT);
	return 0;
}

int recycleDirentries(struct DPB *b,struct dirpointer *start,byte count)
//look for count free Direntries and set dp according, we do not extend a dir
//return 1 if not found. return 2 if not found but empty entries at end of cluster.
{
	void *data;
	struct direntry *dir;
	struct FATfile f;
	struct dirpointer dp;
	int eof;
	word i,entries_per_sector;
	byte n;

	dp=*start;
	data=(direntry*)malloc(b->bytes_per_sector);
	if (data==NULL) {
		printf(msg[39]);

		exit(1);
	}
	entries_per_sector=b->bytes_per_sector/32;
	if ((b->fat_type==FAT32) && (dp.cluster==0)) dp.cluster=b->first_root_cluster;
	f.cluster=dp.cluster;

	if (f.cluster==0) //root?
	{
		f.sector=b->first_root_sector;
	}
	else f.sector=firstSectorOfCluster(b,dp.cluster);

	n=count;
	do {
		do {
			if (readSector(b->drive,f.sector+dp.relsector,1,data)) {
				printf(msg[40]);
				free(data);

				exit(1);
			};
			dir=&((struct direntry*)data)[dp.entry_no];
			for (i=dp.entry_no;i<entries_per_sector;i++) {
				if ((dir->name[0]==0)              //subdir end or
					|| ((byte)dir->name[0]==0xe5)) //deleted
				{
					if (n==count) *start=dp;
					n--;
				}
				else n=count;
				if (n==0) {
					free(data);
					return 0; //found!
				}
				dp.entry_no++;
				dir++;
			} //for each entry
			dp.entry_no=0;
			dp.relsector++;
		} while (((f.cluster!=0) && (dp.relsector<b->sectors_per_cluster)) ||
			((f.cluster==0) && (dp.relsector<b->no_root_sectors)));
		if (f.cluster==0) {
			free(data);
			return 1; //root:not found
		}
		eof=getNextCluster(b,&f);
		if (eof==1) {
			free(data);
			if (n!=count) return 2; //some free entries at the end but not enough
			else return 1; //not enough free space entries found
		}
		dp.relsector=0;
		dp.cluster=f.cluster;
	} while (eof!=1);
	//--this code is never executed, unless you insert a break command above
	free(data);
	if (n!=count) return 2; //some free entries at the end but not enough
	else return 1; //not enough free space entries found
	//--
}

int insertDirentry(struct DPB *b,struct longdirentry *dir,struct longdirentry *f, int protection)
//function appends new direntry on a directory. We recycle deleted direntries.
//protection=PROTECT_ALL: no files are overwritten
//protection=PROTECT_RO: read-only files are not overwritten, normal files are
//protection=PROTECT_SYS: system and hidden files are not overwritten, r-o and normal files are
//protection=PROTECT_NO: all files are overwritten
//returns 0 if successfull, !=0 otherwise
//        1 if file is not writable (i.e. a directory, label)
//        2 if file is system or hidden
//        3 if file is read only
//        4 if file is normal
//        NOTE: dir is used as a DIRPOINTER, only dir->start_cluster and dir->length
//        should be better a dirpointer

{
	struct dirpointer *dp;
	struct longdirentry F;
	struct direntry *pDir;
	struct lfnentry *pLfn;
	byte no_lfn,lfn_len,lfn_pos,lfn_cnt;
	int finished=0;
	int append;
	int excludeflags;
	dword sector,sector2;
	struct FATfile fatf;

	dp=new dirpointer;
	if (dp==NULL) {
		printf(msg[41]);

		exit(1);
	}
	dp->cluster=dir->start_cluster;
	dp->relsector=0;
	dp->entry_no=0;
	dp->countfile=COUNTFILE_START;
	dp->countdir=COUNTDIR_START;
	dp->length=dir->length;
	F=FindMask(b,f->name,dp,1+2+4);        // exact shortname match
	if (F.name[0]!=0) //overwrite Files
	{
		if (protection==PROTECT_ALL) {
			delete dp;
			if ((F.atrib & F_READ_ONLY) !=0) return 3;
			if ((F.atrib & (F_HIDDEN|F_SYSTEM)) !=0) return 2;
			return 4;
		}
		excludeflags=(F_DIR|F_LABEL);
		if (protection==PROTECT_RO) excludeflags|=F_READ_ONLY;
		if (protection==PROTECT_SYS) excludeflags|=F_READ_ONLY|F_HIDDEN|F_SYSTEM;
		if ((F.atrib & excludeflags) !=0) // do not overwrite protected files
		{
			delete dp;
			if ((F.atrib & F_READ_ONLY) !=0) return 3;
			return 2;
		}
		freeClusterChain(b,&F);
		dp->entry_no--;
		deleteDirentry(b,dir,dp,&F);
		F=FindMask(b,f->name,dp,1+2+4); //seek to end of dir
		if (F.name[0]!=0) //another equal entry? strange!
		{
			printf(msg[42]);
			delete dp;

			exit(1);
		}
	}
	if ((strncmp(f->dosname,".          ",11)!=0) && (strncmp(f->dosname,"..         ",11)!=0)) {
		makeAlias(b,dir,f);
	}
	if (f->uniname[0]==0) //no lfn
		no_lfn=0;
	else //has lfn
	{
		lfn_len=Wstrlen(f->uniname);
		no_lfn=lfn_len/CHR_PER_LFN; //how many lfns are used?
		if (Wstrlen(f->uniname)%CHR_PER_LFN!=0) no_lfn++;

	}

	dp->cluster=dir->start_cluster;
	dp->relsector=0;
	dp->entry_no=0;
	dp->countdir=COUNTDIR_START;
	dp->countfile=COUNTFILE_START;
	dp->length=dir->length;
	append=recycleDirentries(b,dp,no_lfn+1);
	if ((append!=0) && (dp->cluster==0)) {
		printf(msg[43],b->no_root_entries);
		delete dp;

		exit(1);
	}
	//dp points to beginning of free space
	pDir=(struct direntry*)malloc(b->bytes_per_sector);
	pLfn=(struct lfnentry*)pDir;
	if (pDir==NULL) {
		printf(msg[44]);
		delete dp;

		exit(1);
	}
	lfn_cnt=no_lfn; //count them
	sector=0;  //invalid
	do {
		if ((dp->entry_no>=b->dir_entries_per_sector) || (append!=0)) //sector boundary?
		{
			if (append!=2) //remember start point?
			{
				dp->entry_no=0;
				dp->relsector++;
			}
			if (((dp->relsector>=b->sectors_per_cluster) && (dp->cluster!=0)) || (append!=0))//cluster boundary?
			{
				if (append!=2) dp->relsector=0;
				fatf.cluster=dp->cluster;
				fatf.sector=firstSectorOfCluster(b,dp->cluster);
				if ((append!=0) || (getNextCluster(b,&fatf)==1))        //if there is no next cluster
					if (extendFile(b,&fatf,1,W_DIR)==0)   //allocate one
					{
					printf(msg[45]);
					delete dp;
					free(pDir);

					exit(1);
					}
					if (append!=2) dp->cluster=fatf.cluster;
				append=0;
			}
		}
		sector2=firstSectorOfCluster(b,dp->cluster)+dp->relsector;
		if (sector!=sector2) //still in the same sector?
		{
			if (sector!=0) {
				writeSector(b->drive,sector,1,pDir,W_DIR);
			}
			readSector(b->drive,sector2,1,pDir);
			sector=sector2;
		}
		if (lfn_cnt!=0) //write lfn
		{
			pLfn[dp->entry_no].count=lfn_cnt;
			if (lfn_cnt==no_lfn)    //first entry is special
			{
				pLfn[dp->entry_no].count|=0x40; //mark first entry
				int tail=lfn_len%(CHR_PER_LFN);
				if (tail==0) tail=CHR_PER_LFN;
				lfn_pos=lfn_len-tail;
				for (int s=1;s<CHR_PER_LFN-tail;s++) f->uniname[s+lfn_len]=0xffff; //pad with ff AFTER 0
				unincpy(pLfn[dp->entry_no].name3,&f->uniname[lfn_pos+11],2);
				unincpy(pLfn[dp->entry_no].name2,&f->uniname[lfn_pos+5],6);
				unincpy(pLfn[dp->entry_no].name1,&f->uniname[lfn_pos],5);
			}
			else {
				if (lfn_len>=CHR_PER_LFN) lfn_pos=lfn_len-CHR_PER_LFN; //first letter to store
				else lfn_pos=0;
				unincpy(pLfn[dp->entry_no].name3,&f->uniname[lfn_pos+11],2);
				unincpy(pLfn[dp->entry_no].name2,&f->uniname[lfn_pos+5],6);
				unincpy(pLfn[dp->entry_no].name1,&f->uniname[lfn_pos],5);
			}
			pLfn[dp->entry_no].atrib=F_LFN;
			pLfn[dp->entry_no].ftype=0;
			pLfn[dp->entry_no].checksum=makeChecksum(f->dosname);
			pLfn[dp->entry_no].reserved=0x0000;
			lfn_len=lfn_pos;
			lfn_cnt--;
		}
		else //write shortentry
		{
			strncpy(pDir[dp->entry_no].name,f->dosname,8);
			strncpy(pDir[dp->entry_no].ext,f->dosext,3);
			pDir[dp->entry_no].atrib=f->atrib;
			strncpy(pDir[dp->entry_no].reserved,f->reserved,8);
			pDir[dp->entry_no].high_cluster=(word)((f->start_cluster & 0xffff0000L) >> 16);
			pDir[dp->entry_no].time=f->time;
			pDir[dp->entry_no].date=f->date;
			pDir[dp->entry_no].start_cluster=(word)(f->start_cluster & 0x0000ffffL);
			pDir[dp->entry_no].length=f->length;
			finished=1;
			writeSector(b->drive,sector,1,pDir,W_DIR);
		}
		dp->entry_no++;
	} while (finished!=1);
	delete dp;
	free(pDir);
	return 0;
}

int strlast(char *s,char c) //find last occurence of c in s, return position of c or 0xffff
{
	int l,j;
	l=0xffff;
	for (j=0;s[j]!=0;j++)
		if (s[j]==c) l=j;
	return l;
}

int winChar(char c)
//1: if is a win allowed character , 0 else
{
	if ((c==43) || (c==44) || (c==59) || (c==61) ||
		(c==91) || (c==93)) return 1;
	else return 0;
}

// same as dosChar but more restrictive checking
// to make it the thing codepage independent

int dosCharBase(unsigned char c)
{
	if (c >= 127)                         // only us charset !
	   return 0;

	if (   (c > 32)                       //exclude 0 - blank
	    && (!(c>=97 && c<=122))           //exclude a - z
	    && (c!= 34)                       //exclude "
		 && (!(c>=42 && c<=44))            //exclude * - ,
	    && (c!=46)                        //exclude .
		 && (c!=47)                        //eclude  /
	    && (!(c>=58 && c<=63))            //exclude : - ?
	    && (!(c>=91 && c<=93))            //exclude [\]
	    && (c!=124)                       //exclude |
	   )
	   return 1;                          // valid char
	else
	   return 0;
}


int dosChar(unsigned char c)
{
	if ((c<128)                    //graphical
		&& ((c>123) || (c<94))        //^-a-z-{
		&& ((c>90) || (c<64))         //@-Z
		&& ((c>57) || (c<48))         //0-9
		&& (c!=33)                    //!
		&& ((c<35) || (c>41))         //#-)
		&& (c!=45)                    //-
		&& ((c<125) || (c>126)))       //} and ~
		return 0;                    //invalid char
	else
		return 1;
}

int dosStr(char *s)
{
	char *c;
	for (c=s;*c!=0;c++) {
		if ((dosChar(*c)==0) && (*c!='.')) return 0;
	}
	return 1; //all valid characters
}

int dosStrBase(const char *s)
{
	const char *c;
	for (c=s;*c!=0;c++) {
		if ((dosCharBase(*c)==0) && (*c!='.')) return 0;
	}
	return 1; //all valid characters
}

// check unicode string for valid chars
int uniStrBase(word *ws)
{
	word *w;
	for (w=ws;*w!=0;w++) {
		if (*w > 255) return 0;

		if ((dosCharBase((byte)*w)==0) && ((byte)(*w)!='.')) return 0;
	}
	return 1; //all valid characters
}


void convertInput(char *s)
//convert Windows ANSI to Codepage 850
{
	char *c;
	for (c=s;*c!=0;c++)
		*c=win850[(unsigned char)*c];
}


//creates a pseudo ascii string from unicode, all invalid chars are translated
//valid chars are from 0x20-7E, all other will be converted to _

void Uni2DOSString(byte *out, word *win)
{
   word *w;
   byte *o;
   for (w=win, o=out;*w!=0;w++,o++) {
       if (*w < 32 || *w > 126)
	  *o='_';
       else
	  *o=(byte)(*w);
	}
   *o = 0;
}


// creates unicode from ascii, no conversion
void Ascii2UniString(word *to, byte *from)
{

  for (; (*to = (word)*from) != 0; ++from, ++to);
}

// compare unicode longfilename with dos shortfilename
// returns 0 if exactmatch
RealLongname(struct longdirentry *plfnd)
{
	  word unishort[16];

     Ascii2UniString(unishort, (byte *)shortname(plfnd));
     return Wstrcmp((const short *)unishort, (const short *)plfnd->uniname);
}


//creates a pseudo ascii string, all invalid chars are translated
//valid chars are from 0x20-7E, all other will be converted to _

void Ascii2DOSString(byte *out, byte *in)
{
   byte *w;
   byte *o;

   for (w=in, o=out;*w!=0;w++,o++) {
       if (*w < 32 || *w > 126)
	  *o='_';
       else
	  *o=(byte)(*w);
   }
   *o = 0;
}


// makeAlias parameter from dir: only start_cluster and length
// f contains name info in dosname and name
// note: if there is already a dosname that is unique we will keep it
// to force a new alias set f->dosname[0] = 0;

//#if 0
void makeAlias(const struct DPB *b, const struct longdirentry *dir, struct longdirentry *f)
{
	unsigned int dot,len,i,j,tilda;
	char number[7],mask[13];
	struct dirpointer dp;
	struct longdirentry r;


	dp.cluster=dir->start_cluster;
	dp.relsector=0;
	dp.entry_no=0;
	dp.countfile=COUNTFILE_START;
	dp.countdir=COUNTDIR_START;
	dp.length=dir->length;
	if (f->dosname[0]!=0) {            // is there already a shortname ?
		strcpy(mask,shortname(f));
		if (mask[0]!=0) {
			r=FindMask(b,mask,&dp,1+2); // find shortname exact
			if (r.name[0]==0) // short name is already unique, so we keep it
			{
				f->checksum=makeChecksum(f->dosname);
				return;
			}
		}
	}

					 // name is not unique or no dosname given, make new alias
	dot=strlast(f->name,'.'); //last dot
	len=strlen(f->name);
	if  ( (((dot == 0xffff) && (len<=8)) ^      //no dot Xor
		((strpos(f->name,len,'.')==dot) &&  //only one dot and
		(dot<=8) &&   (len-dot<=4))  ) &&   //8.3 AND
		(dosStr(f->name)==1          )   )  //all chars valid

	{
		if (dot==0xffff) {
			for (i=0;i<len;i++) f->dosname[i]=asciiUpcase[(unsigned char)f->name[i]];
			for (i=len;i<8;i++) f->dosname[i]=' ';
			for (i=0;i<3;i++) f->dosext[i]=' ';
		}
		else {
			for (i=0;i!=dot;i++) f->dosname[i]=asciiUpcase[(unsigned char)f->name[i]];
			for (i=dot;i<8;i++) f->dosname[i]=' ';      //padding
			for (i=0;dot+i+1<len;i++) f->dosext[i]=asciiUpcase[(unsigned char)f->name[dot+1+i]];
			for (i=len-dot-1;i<3;i++) f->dosext[i]=' '; //padding
			len=dot;
		}
		tilda=1; //next number if not unique
	}
	else //make a short alias
	{
		//name
		for (i=j=0;(f->name[j]!='\0') && (i<6) && (j<=dot);j++) {
			if (dosChar(f->name[j])==1) {
				f->dosname[i]=asciiUpcase[(unsigned char)f->name[j]];
				i++;
			}
			else if (winChar(f->name[j])==1) {
				f->dosname[i]='_';
				i++;
			}
		}
		len=i;
		f->dosname[i]=126;i++; //~1
		f->dosname[i]='1';i++;
		//padding
		for (;i<8;i++) f->dosname[i]=' ';
		//extension
		if (dot!=0xffff) {
			j=dot+1;
			for (i=0;(i<3) && (f->name[j]!=0);j++) {
				if (dosChar(f->name[j])==1) {
					f->dosext[i]=asciiUpcase[(unsigned char)f->name[j]];
					i++;
				}
				else if (winChar(f->name[j])==1) {
					f->dosext[i]='_';
					i++;
				}
			}
			//padding
			for (;i<3;i++) f->dosext[i]=' ';
		}
		else strncpy(f->dosext,"   ",3);
		tilda=2; //next number if not unique
	}
	//len=first char after name (<=6)
	//is it unique?

	do {
		dp.cluster=dir->start_cluster;
		dp.relsector=0;
		dp.entry_no=0;
		dp.countfile=COUNTFILE_START;
		dp.countdir=COUNTDIR_START;
		dp.length=dir->length;
		strcpy(mask,shortname(f));
		r=FindMask(b,mask,&dp,1+2);
		if (r.name[0]!=0) //short name does already exist
		{
			if (strcmp(r.name, f->name)!=0) //long names are the same? n: create id alias
			{
				itoa(tilda,number,10);
				while (len+strlen(number)>=8) len--;
				f->dosname[len]=126; //~
				for (i=0;number[i]!=0;i++) f->dosname[len+1+i]=number[i];
				tilda++;
			}
			else //y: overwrite
			{
				strncpy(f->dosname,r.dosname,8);
				strncpy(f->dosext,r.dosext,3);
				break;
			}
		}
	} while (r.name[0]!=0); //until unique
	f->checksum=makeChecksum(f->dosname);
}

//#endif

#if 0
void makeAlias(const struct DPB *b, const struct longdirentry *dir, struct longdirentry *f)
{
	unsigned int dot,len,i,j,tilda,lmatch;
	char number[7],mask[13];
	struct dirpointer dp;
	struct longdirentry r;


	dp.cluster=dir->start_cluster;
	dp.relsector=0;
	dp.entry_no=0;
	dp.countfile=COUNTFILE_START;
	dp.countdir=COUNTDIR_START;
	dp.length=dir->length;
	if (f->dosname[0]!=0) {            // is there already a shortname ?
		strcpy(mask,shortname(f));
		if (mask[0]!=0) {
			r=FindMask(b,mask,&dp,1+2); // find shortname exact
			if (r.name[0]==0) // short name is already unique, so we keep it
			{
				f->checksum=makeChecksum(f->dosname);
				return;
			}
		}
	}

			       // name is not unique or no dosname given, make new alias
	dot=strlast(f->name,'.'); //last dot
	len=strlen(f->name);
	if  ( (((dot == 0xffff) && (len<=8)) ^        // no dot Xor
		((strpos(f->name,len,'.') == dot) &&  // only one dot and
		(dot<=8) &&   (len-dot<=4))  ) &&     // 8.3 AND
		(dosStrBase(f->name)==1      )  )

	{
		if (dot==0xffff) {                 //all chars valid
			for (i=0;i<len;i++) f->dosname[i]=asciiUpcase[(unsigned char)f->name[i]];
			for (i=len;i<8;i++) f->dosname[i]=' ';
			for (i=0;i<3;i++) f->dosext[i]=' ';
		}
		else {
			for (i=0;i!=dot;i++) f->dosname[i]=asciiUpcase[(unsigned char)f->name[i]];
			for (i=dot;i<8;i++) f->dosname[i]=' ';      //padding
			for (i=0;dot+i+1<len;i++) f->dosext[i]=asciiUpcase[(unsigned char)f->name[dot+1+i]];
			for (i=len-dot-1;i<3;i++) f->dosext[i]=' '; //padding
			len=dot;
		}
		tilda=1; //next number if not unique
	}
	else //make a short alias
	{
		//name
		for (i=j=0;(f->name[j]!='\0') && (i<6) && (j<=dot);j++) {
			if (dosCharBase(f->name[j])==1) {
				f->dosname[i]=asciiUpcase[(unsigned char)f->name[j]];
				i++;
			}
			else if (winChar(f->name[j])==1) {
				f->dosname[i]='_';
				i++;
			}
		}
		len=i;
		f->dosname[i]=126;i++; //~1
		f->dosname[i]='1';i++;
		//padding
		for (;i<8;i++) f->dosname[i]=' ';
		//extension
		if (dot!=0xffff) {
			j=dot+1;
			for (i=0;(i<3) && (f->name[j]!=0);j++) {
				if (dosCharBase(f->name[j])==1) {
					f->dosext[i]=asciiUpcase[(unsigned char)f->name[j]];
					i++;
				}
				else if (winChar(f->name[j])==1) {
					f->dosext[i]='_';
					i++;
				}
			}
			//padding
			for (;i<3;i++) f->dosext[i]=' ';
		}
		else strncpy(f->dosext,"   ",3);
		tilda=2; //next number if not unique
	}
	//len=first char after name (<=6)
	//is it unique?

	do {
		dp.cluster=dir->start_cluster;
		dp.relsector=0;
		dp.entry_no=0;
		dp.countfile=COUNTFILE_START;
		dp.countdir=COUNTDIR_START;
		dp.length=dir->length;
		strcpy(mask,shortname(f));
		r=FindMask(b,mask,&dp,1+2);
		if (r.name[0]!=0)          // short name does already exist
		{
#if 0

						// long names are the same? n: create id alias
						// make unicode compare if unicode exist
			if (r.uniname[0] != 0 && f->uniname[0] != 0)
				  lmatch = !Wstrcmp((const short *)r.uniname, (const short *)f->uniname);
			else
				  lmatch = !strcmp(r.name, f->name);

			if (!lmatch)
			{
#endif
				itoa(tilda,number,10);
				while (len+strlen(number)>=8) len--;
				f->dosname[len]=126; //~
				for (i=0;number[i]!=0;i++) f->dosname[len+1+i]=number[i];
				tilda++;
#if 0

			}
			else //longmatch yes: overwrite
			{
				strncpy(f->dosname,r.dosname,8);
				strncpy(f->dosext,r.dosext,3);
				break;
			}
#endif
		}
	} while (r.name[0]!=0); //until unique
	f->checksum=makeChecksum(f->dosname);
}
#endif



word *unichr(word *s, word c)
//Find first occurence of a charcter in an unicode string or return NULL
{
	for (word *p=s;*p!=0;p++) {
		if (*p==c) return p;
	}
	return NULL;
}




void setDate(char* date, struct longdirentry *f) {
 word year=((date[6]-'0')*10+(date[7]-'0'));
 if (year<80) year=(2000+year)-1980;
 else year=(1900+year)-1980;
 f->date=(year << 9) | (((date[0]-'0')*10+(date[1]-'0')) << 5) | ((date[3]-'0')*10+(date[4]-'0'));
}

void setTime(char* time, struct longdirentry *f) {
 f->time=(((time[0]-'0')*10+(time[1]-'0')) << 11) | (((time[3]-'0')*10+(time[4]-'0')) << 5) |
		  (((time[6]-'0')*10+(time[7]-'0')) >> 1);
}

void setCurrentDateTime(struct longdirentry *f) {
 char time[10];
 char date[10];
 _strtime(time);
 _strdate(date);
 setDate(date, f);
 setTime(time, f);
}

int initDirectory(struct DPB *b, struct longdirentry *parent, struct longdirentry *dir) {
 struct longdirentry g;

 g.name[0]='.';
 g.name[1]=0;
 g.uniname[0]=0;
 for (int i=0;i<11;i++) g.dosname[i]=' ';
 g.dosname[0]='.';
 g.atrib=F_DIR;
 g.ftype=0;
 g.checksum=0;
 setCurrentDateTime(&g);
 g.start_cluster=dir->start_cluster;
 g.length=0;
 for (i=0;i<8;i++) g.reserved[i]=0;
 if (insertDirentry(b,dir,&g,PROTECT_ALL)!=0)
 {
  return 1;
 }
 g.name[1]='.';
 g.name[2]=0;
 g.dosname[1]='.';
 if (parent->start_cluster!=b->first_root_cluster) {
  g.start_cluster=parent->start_cluster;
 }
 else {
  g.start_cluster=0; //points to root
 }
 if (insertDirentry(b,dir,&g,PROTECT_ALL)!=0)
 {
  return 1;
 }
 return 0;
}

int makeDirectory(struct DPB *b, struct longdirentry *parent, longdirentry *f) {
 struct FATfile fat;

 fat.cluster=getNextAvailClusterNo(b);
 fat.sector=firstSectorOfCluster(b,fat.cluster);
 f->start_cluster=fat.cluster;

 if (extendFile(b,&fat,1,W_DIR)!=1)
 {
  return 1;
 }

 if (insertDirentry(b, parent, f, PROTECT_ALL)!=0)
 {
  return 2;
 }

 if (initDirectory(b, parent, f)!=0) {
  return 3;
 }
 return 0;
}

int makeDirectory(struct DPB *b, struct longdirentry *parent, char *name, longdirentry *f) {
 int trash;

 strcpy(f->name,name);
 f->uniname[0]=0;
 xlate_to_uni((unsigned char*) name, strlen(name),(unsigned char*) f->uniname, &trash, '\\');
 f->uniname[trash] = 0;
 f->dosname[0]=0;
 f->dosext[0]=0;
 f->atrib=F_DIR;
 f->ftype=0;
 f->checksum=0;
 setCurrentDateTime(f);
 f->length=0;
 strnset(f->reserved,0,8);
 return makeDirectory(b, parent, f);
}


